@@ -1,12 +1,12 @@
 _uftrace () {
-    local cur prev subcmd options uftrace_comp
+    local cur prev subcmds options uftrace_comp
 
     cur=${COMP_WORDS[COMP_CWORD]}
     prev=${COMP_WORDS[COMP_CWORD - 1]}
 
     COMPREPLY=()
 
-    subcmds='record replay report live dump graph info recv script'
+    subcmds='record replay report live dump graph info recv script tui'
     options=$(uftrace -? | awk '$1 ~ /--[a-z]/ { split($1, r, "="); print r[1] } \
                                 $2 ~ /--[a-z]/ { split($2, r, "="); print r[1] }')
     demangle='full simple no'
