@@ -1,11 +1,12 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <alloca.h>
+#include <string.h>
 
 int foo(int c)
 {
 	char *ptr = alloca(c);
-	snprintf(ptr, c, "%s", "hello world\n");
+	strncpy(ptr, "hello world\n", c);
 	return c;
 }
 
