@@ -23,13 +23,13 @@ class TestCase(TestBase):
 """, sort='simple')
 
     def pre(self):
-        record_cmd = '%s record -d %s %s' % (TestBase.ftrace, TDIR, 't-namespace')
+        record_cmd = '%s record -d %s %s' % (TestBase.uftrace_cmd, TDIR, 't-namespace')
         sp.call(record_cmd.split())
         return TestBase.TEST_SUCCESS
 
     def runcmd(self):
         return '%s replay -d %s --disable -N "ns2.*" -T "operator new@trace-on" -T "malloc@traceoff"' % \
-            (TestBase.ftrace, TDIR)
+            (TestBase.uftrace_cmd, TDIR)
 
     def post(self, ret):
         sp.call(['rm', '-rf', TDIR])
