@@ -24,4 +24,4 @@ class TestCase(TestBase):
                                       ['libabc_test_lib.so'])
 
     def runcmd(self):
-        return '%s --force --no-libcall -A ^lib@libabc_test,arg1 %s' % (TestBase.ftrace, 't-' + self.name)
+        return '%s --force --no-libcall -A ^lib@libabc_test,arg1 %s' % (TestBase.uftrace_cmd, 't-' + self.name)
