@@ -23,7 +23,7 @@ class TestCase(TestBase):
 """, sort='graph')
 
     def pre(self):
-        record_cmd = '%s record -d %s %s' % (TestBase.ftrace, TDIR, 't-' + self.name)
+        record_cmd = '%s record -d %s %s' % (TestBase.uftrace_cmd, TDIR, 't-' + self.name)
         sp.call(record_cmd.split())
         return TestBase.TEST_SUCCESS
 
@@ -39,7 +39,7 @@ class TestCase(TestBase):
                 pass
         if t == 0:
             return 'FAILED TO FIND TID'
-        return '%s graph -d %s --tid %d %s' % (TestBase.ftrace, TDIR, t, FUNC)
+        return '%s graph -d %s --tid %d %s' % (TestBase.uftrace_cmd, TDIR, t, FUNC)
 
     def post(self, ret):
         sp.call(['rm', '-rf', TDIR])
