@@ -2,7 +2,7 @@
 
 from runtest import TestBase
 import subprocess as sp
-import os
+import os, re
 
 TDIR  = 'xxx'
 TDIR2 = 'yyy'
@@ -19,7 +19,6 @@ class TestCase(TestBase):
   89.018 us [18343] |   } /* fopen */
             [18343] |   fclose() {
   10.781 us [18343] |     sys_close();
-  21.980 us [18343] |     exit_to_usermode_loop();
   37.325 us [18343] |   } /* fclose */
  128.387 us [18343] | } /* main */
 """)
@@ -32,16 +31,21 @@ class TestCase(TestBase):
         if os.path.exists('/.dockerenv'):
             return TestBase.TEST_SKIP
 
-        recv_cmd = '%s recv -d %s' % (TestBase.ftrace, TDIR)
+        uftrace = TestBase.uftrace_cmd
+        program = 't-' + self.name
+
+        recv_cmd = '%s recv -d %s' % (uftrace, TDIR)
         self.recv_p = sp.Popen(recv_cmd.split())
 
-        record_cmd = '%s record -H %s -k -N %s@kernel -d %s %s' % \
-                     (TestBase.ftrace, 'localhost', 'smp_irq_work_interrupt', TDIR2, 't-' + self.name)
+        argument  = '-H %s -k -d %s' % ('localhost', TDIR2)
+        argument += ' -N %s@kernel' % '_*do_page_fault'
+
+        record_cmd = '%s record %s %s' % (uftrace, argument, program)
         sp.call(record_cmd.split())
         return TestBase.TEST_SUCCESS
 
     def runcmd(self):
-        return '%s replay -d %s' % (TestBase.ftrace, os.path.join(TDIR, TDIR2))
+        return '%s replay -d %s' % (TestBase.uftrace_cmd, os.path.join(TDIR, TDIR2))
 
     def post(self, ret):
         self.recv_p.terminate()
@@ -49,4 +53,12 @@ class TestCase(TestBase):
         return ret
 
     def fixup(self, cflags, result):
-        return result.replace('sys_open', 'sys_openat')
+        uname = os.uname()
+
+        # Linux v4.17 (x86_64) changed syscall routines
+        major, minor, release = uname[2].split('.')
+        if uname[0] == 'Linux' and uname[4] == 'x86_64' and \
+           int(major) >= 4 and int(minor) >= 17:
+            return re.sub('sys_[a-zA-Z0-9_]+', 'do_syscall_64', result)
+        else:
+            return result.replace('sys_open', 'sys_openat')
