@@ -27,10 +27,10 @@ class TestCase(TestBase):
 
     def pre(self):
         self.file_p = open(TMPF, 'w+')
-        recv_cmd = TestBase.ftrace.split() + ['recv', '-d', TDIR, '--run-cmd', TestBase.ftrace + ' replay']
+        recv_cmd = TestBase.uftrace_cmd.split() + ['recv', '-d', TDIR, '--run-cmd', TestBase.uftrace_cmd + ' replay']
         self.recv_p = sp.Popen(recv_cmd, stdout=self.file_p, stderr=self.file_p)
 
-        record_cmd = '%s record -H %s %s' % (TestBase.ftrace, 'localhost', 't-' + self.name)
+        record_cmd = '%s record -H %s %s' % (TestBase.uftrace_cmd, 'localhost', 't-' + self.name)
         sp.call(record_cmd.split())
         return TestBase.TEST_SUCCESS
 
