@@ -21,13 +21,18 @@ class TestCase(TestBase):
 """, sort='report')
 
     def pre(self):
+        if not TestBase.check_dependency(self, 'perf_context_switch'):
+            return TestBase.TEST_SKIP
+        if not TestBase.check_perf_paranoid(self):
+            return TestBase.TEST_SKIP
+
         options = '-d %s -E %s' % (TDIR, 'linux:schedule')
-        record_cmd = '%s record %s %s' % (TestBase.ftrace, options, 't-' + self.name)
+        record_cmd = '%s record %s %s' % (TestBase.uftrace_cmd, options, 't-' + self.name)
         sp.call(record_cmd.split())
         return TestBase.TEST_SUCCESS
 
     def runcmd(self):
-        return '%s report -d %s' % (TestBase.ftrace, TDIR)
+        return '%s report -d %s' % (TestBase.uftrace_cmd.split()[0], TDIR)
 
     def post(self, ret):
         sp.call(['rm', '-rf', TDIR])
